#! /usr/bin/perl
package PortParser;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(parse_ports randomize_ports);

#use strict;

#sub parse_ports;
#sub test_ports;
#sub randomize_ports;

#test_ports($ARGV[0]);

sub parse_ports {
    my $line = shift;
    my @temp_ports = split(/,/, $line);

    my %port_check;  # associative array used to ensure that a port number is
                     # not added more than once

    my @ports;       # used to store the final processed port numbers

    my @ports_pair;  # used to store two port numbers in x-y format, e.g. 3-6

    my $temp_port;
    foreach $temp_port (@temp_ports) {
        # Test if this line is an "x-y" type, e.g. 3-6
        @ports_pair = split(/-/, $temp_port);

        # If it's not an "x-y" type, just add this port to the ports array
        # and go to the next loop
        if ($#ports_pair == 0) {
            if (!exists($port_check{$temp_port})) {
                push(@ports, $temp_port);
                $port_check{$temp_port} = 1;
            }
            next;
        }

        # If the first port number is larger than the other, swap them around
        if ($ports_pair[0] > $ports_pair[1]) {
            $ports_pair[0] ^= $ports_pair[1];
            $ports_pair[1] ^= $ports_pair[0];
            $ports_pair[0] ^= $ports_pair[1];
        }

        # Go through the numbers from the first port number to the second
        # port number and add them to the ports array.
        # e.g. if it's 3-6, then add ports 3, 4, 5, 6 to the ports array
        for (my $i = $ports_pair[0]; $i <= $ports_pair[1]; $i++) {
            if (!exists($port_check{$i})) {
                push(@ports, $i);
                $port_check{$temp_port} = 1;
            }
        }
    }

    # Sort the port numbers numerically
    @ports = sort {$a <=> $b} (@ports);

    return @ports;
}

sub test_ports {
    my $line = shift || "3-6,8,9,12-30";
    my @ports;
    
    @ports = parse_ports($line);
    
    my $port;
    print "$line\n";
    foreach $port (@ports) {
        print "Port $port\n";
    }

    @ports = randomize_ports(join(":", @ports));

    print "After randomizing,\n";
    foreach $port (@ports) {
        print "Port $port\n";
    }
}

# Randomize an array of ports and returns the randomized array.
# Call this function this way:
# randomize_ports(join(":", @ports))
sub randomize_ports {
    my $line = shift;
    my @temp_ports = split(/:/, $line);
    my @ports;
    my %port_check;
    my $num_of_ports = ($#temp_ports + 1);
    my $random_pos;

    for (my $counter = 0; $counter < $num_of_ports; $counter++) {
        $random_pos = int(rand($num_of_ports));
        if (!exists($port_check{$random_pos})) {
            #print "ports[$random_pos] = temp_ports[$counter] = $temp_ports[$counter]\n";
            $ports[$random_pos] = $temp_ports[$counter];
            $port_check{$random_pos} = 1;
        } else {
            # quick hack.
            $counter--;
        }
    }

    return @ports;
}
