eval 'exec perl -x $0 ${1+"$@"}' # -*-perl-*-
  if 0;
#!perl
#
# SIDEN - a distributed network discovery tool
# Copyright (C) 2000 Lawrence Teo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# $Id: siden.pl,v 1.2 2000/08/07 16:08:08 steelx Exp steelx $

use strict;
use vars '$sh';
use vars qw($opt_s $opt_p $opt_a $opt_d $opt_c $opt_o
            $opt_r $opt_v $opt_V $opt_D $opt_h);
use Net::Inet;
use Net::TCP;
use Digest::MD5 qw(md5_hex);
use Getopt::Std;
use POSIX;

use SidenMeta;
use PortParser;     # Use the PortParser class
use Scan;           # Use the Scan class
use Agent;          # Use the Agent class

sub usage;
sub process_results;
sub logmsg;
sub logdebug;

my $DEBUG = 0;
my $VERBOSE = 0;


# Display the usage if there are no arguments
if ($#ARGV == -1) {
    usage;
}

# Process the options
usage if (!getopts("s:p:a:d:o:D:rvVhc"));

my $output_file;
$output_file = $opt_o || "&STDOUT";

open(OUTPUT, ">>$output_file") or die("Cannot open $output_file\n");

logmsg "SIDEN version $SidenMeta::VERSION\n";
logmsg "Copyright (c) $SidenMeta::YEAR by $SidenMeta::AUTHOR <$SidenMeta::EMAIL>.\n\n";

if (defined($opt_h)) {
    usage();
}

if (defined($opt_V)) {
    print "siden Version $SidenMeta::VERSION\n";
    exit(1);
}
# Parameters
my @hosts;
my @decoys;
my $scan_type;
my $delay;
my $randomize;
my $concurrent;
my $services_file_found = 1;
my %services; # Hash to map a port number to a service name
my $service;
my $port_number;

# Read the services file
if (!(-r $SidenMeta::SERVICES)) {
    print "Can't read $SidenMeta::SERVICES\n" if ($DEBUG >= 1);
    $services_file_found = 0;
}

if ($services_file_found == 1) {
    print "Opening $SidenMeta::SERVICES\n" if ($DEBUG >= 1);
    open(SVCFILE, "<$SidenMeta::SERVICES") or print("Can't open services\n");
    while (<SVCFILE>) {
        chomp;
        undef($service);
        undef($port_number);
        ($service, $port_number) = $_ =~ /^(\S+)\s+(\d+)\/tcp/;
        if ((defined($service)) & (defined($port_number))) {
            print "Port $port_number is $service\n" if ($DEBUG >= 2);
            $services{$port_number} = $service;
        }
    }
    close(SVCFILE);
}

# Process the hosts
my $arg;
foreach $arg (@ARGV) {
    push(@hosts, $arg);
}

# Print debugging info if required
if ($DEBUG >= 2) {
    if ($opt_s eq "S") {
        print "SYN SCAN\n";
    } elsif ($opt_s eq "T") {
        print "FULL CONNECT scan\n";
        $scan_type = "T";
    }
    print "agents $opt_a\n";
    print "delay $opt_d\n";
    print "decoys $opt_D\n";
    print "ports $opt_p\n";

    print "randomize? $opt_r\n";
    print "concurrent? $opt_c\n";

    my $host;
    foreach $host (@hosts) {
        print "host $host\n";
    }
}

$scan_type = $opt_s || "T";
$delay = $opt_d || 100;
$randomize = $opt_r || 0;
$concurrent = $opt_c || 0;

# Break up the -a line into array elements. For example,
# "-aa.b.c.d:3392,w.x.y.z:8833" will break up into
# "a.b.c.d:3392" and "w.x.y.z:8833"
my @temp_agents = split(/,/, $opt_a);
my @agents;

# Further break up the "a.b.c.d:3392" lines and push them into the
# agents array which contains Agent objects
my $agent;
my $newagent;
my $agent_ip;
my $agent_port;

logmsg "Agents\n";
for $agent (@temp_agents) {
    print "agent $agent\n" if ($DEBUG >= 2);
    ($agent_ip, $agent_port) = split(/:/, $agent);
    $newagent = new Agent;
    $newagent->ip($agent_ip);
    $newagent->port($agent_port);
    push(@agents, $newagent);
    logmsg "Agent $agent_ip on port $agent_port\n";
}
logmsg "\n";

if ($DEBUG >= 2) {
    for $agent (@agents) {
    #    print "AGENT " . $agent->ip() . " at port " . $agent->port() . "\n";
        print "AGENT " . $agent->describe_self . "\n";
    }
    print "Number of agents: " .($#agents + 1) . "\n";
}

# Parse the port numbers. Routines parse_ports() and randomize_ports()
# are from the PortParser module.
if (!defined($opt_p)) {
    # Scan ports 1-1024 if no ports are specified
    $opt_p = "1-1024";
}
my @ports = parse_ports($opt_p);

# Put the randomize_ports() statement here for less randomization


if ($DEBUG >= 2) {
    my $port;
    print "Port ";
    for $port (@ports) {
        print "$port,";
    }
    print "\n";
}
print "Number of ports: " . ($#ports + 1) . "\n" if ($DEBUG >= 2);

# A `port block' is (roughly) the number of ports allocated to each agent.
# For example, if there are 10 ports to scan (1..10) and 2 agents, the
# port block will be 5 (meaning each agent scans 5 ports).
# If there are 3 agents and 10 ports, the port block will be 4.
my $port_block = ceil(($#ports + 1) / ($#agents + 1));

print "Port blocks: " . $port_block . "\n" if ($DEBUG >= 2);


# Allocate hosts for agents to scan
my $host;
my @scans;
my $scan;
my $ports_to_scan;
foreach $host (@hosts) {
    my $prev_counter = 0;   # keeps track of the prev counter after each loop
    my $counter = 0;
    my $agent_counter = 0;  # used to count the agents

    # Randomize the ports if we have to
    if ($randomize == 1) {
        @ports = randomize_ports(join(":", @ports));
    }

    foreach $agent (@agents) {
        print "Agent " . $agent->ip() . " scans $host on ports: "
            if ($DEBUG >= 1);

        $agent_counter++;

        # Allocate ports to agents by chunks defined by the port block
        $ports_to_scan = "";
        for ($counter = $prev_counter; $counter < ($port_block + $prev_counter);
             $counter++) {
            if (defined($ports[$counter])) {
                $ports_to_scan .= ($ports[$counter] . ",");
            }
        }

        # If it's the last agent, allocate any remaining port numbers
        if ($agent_counter == ($#agents + 1)) {
            for (; $counter < ($#ports + 1); $counter++) {
                $ports_to_scan .= ($ports[$counter] . ",");
            }
        }

        print "$ports_to_scan\n" if ($DEBUG >= 1);

        $scan = new Scan;
        $scan->agent($agent);
        $scan->type($scan_type);
        $scan->version($SidenMeta::VERSION);
        $scan->delay($delay);
        $scan->host($host);
        $scan->ports_to_scan($ports_to_scan);
        push(@scans, $scan);
 
        $prev_counter = $counter;   # Get ready for the next chunk
    }
}

my %open_ports;
my $results;
my $tmp_file;
my @tmp_files;
my $pid;
foreach $scan (@scans) {
    #my $agent_ip = $scan->agent_ip;
    $agent = $scan->agent;
    print "Agent " . $agent->ip . ":" . $agent->port . "\n" .
          $scan->gen_cmd . "\n\n" if ($DEBUG >= 1);

    if ($concurrent) {
        # Generate a hopefully-really-random temporary filename
        $tmp_file = "/tmp/siden-" . md5_hex(time() . $$ . $agent->ip . 
                    $agent->port . $scan->gen_cmd);

        undef($pid);

        # Fork a new process for the scan
        defined($pid = fork) or die "fork: $!\n";
        if ($pid) {
            # for the parent process
            # Save the temporary file
            push(@tmp_files, $tmp_file);
            next;   # go to the next foreach loop
        }
    }

    $sh = 'Net::TCP'->new($agent->ip, $agent->port);
    $sh->print($scan->gen_cmd . "\n");

    # Get the results from the agent
    $results = readline($sh);
    process_results($results);

    # So long, agent...
    $sh->print("quit");

    if ($concurrent) {
        # Save the results to the temporary file
        open(TMPFILE, ">$tmp_file") or die("Cannot open $tmp_file");
        print TMPFILE $results;
        close(TMPFILE);

        # Bye bye child process
        exit(0);
    }
}

if ($concurrent) {
    # Wait for the results to arrive
    my $results_arrived = 0;
    while (!$results_arrived) {
        $results_arrived = 1;
        foreach $tmp_file (@tmp_files) {
            if (!(-r $tmp_file)) {
                $results_arrived = 0;
            }
        }
        sleep 1;
    }

    # Process the temp files
    foreach $tmp_file (@tmp_files) {
        open(TMPFILE, "<$tmp_file") or die("Can't open $tmp_file\n");
        while (<TMPFILE>) {
            $results = $_;
            process_results($results);
        }
    }
}


# Remove the temp files if not in debug mode
if (!($DEBUG >= 1)) {
    for $tmp_file (@tmp_files) {
        unlink $tmp_file;
    }
}

# Print out the open ports on each host
logmsg "Scan Results\n";
my %port_states = qw(
    + open
    f filtered
);
my $port_state;
my $port;
foreach $host (@hosts) {
    @ports = split(/,/, $open_ports{$host});

    # Sort the ports numerically
    @ports = sort { $a <=> $b } @ports;

    #print "$host: " . $open_ports{$host} . "\n";
    foreach $port (@ports) {
        ($port, $port_state) = $port =~ /^(\d+)([+f])$/;
        logmsg "$host: $port\t$port_states{$port_state}\t\t";

        # print the service if we know it
        if (($services_file_found == 1) & (defined($services{$port}))) {
            logmsg "$services{$port}\n" if (defined($services{$port}));
        } else {
            logmsg "unknown\n";
        }
    }
    logmsg "\n";
}

close(OUTPUT);


# process_results(results)
# WARNING!! USES GLOBAL VARIABLES:
# $host, $tmp_open_ports, $agent, and $open_ports
sub process_results {
    my $results = shift;
    my $tmp_open_ports;

    chomp($results);
    ($host, $tmp_open_ports) = split(/:/, $results);
    
    # if the result is valid, we'll get:
    # target.example.com:1,3,5,7,
    # if the result is invalid, we'll get:
    # target.example.com:error, agent has a problem
    if ($tmp_open_ports =~ /^error,/) {
        $tmp_open_ports =~ s/,/:/;
        logmsg "Agent " . $agent->ip . ": " . $tmp_open_ports . "\n";
    } else {
        $open_ports{$host} .= $tmp_open_ports;
        #$open_ports{$scan->host} .= $tmp_open_ports;
    }
}


sub logmsg {
    print OUTPUT @_;
}

# ======================================================================
# Usage subroutine
# ======================================================================
sub usage {
    print <<USAGEEOF
usage: siden [OPTION] host
OPTION:
    -sT              scan using TCP full connect() scan
    -sS              scan using half-open SYN scan
    -a agent-ip:port use agent listening on specified port of agent-ip
    -p portnums      the ports to use (e.g. -p1-3 or -p1,3,6-7 etc.)
    -d delay         delay for delay secs between scans
  * -D decoy1,decoy2 use decoys
  * -C file          use configuration file "file"
    -o file          log output to file "file"
    -c               scan concurrently
    -r               randomize the scans
  * -v               verbose
    -V               display the version
    -h help          this page

* = not implemented yet
USAGEEOF
;

    exit 1;
}
